<?php

namespace Drupal\advanced_file_destination\Controller;

use Drupal\advanced_file_destination\Service\AdvancedFileDestinationManager;
use Drupal\Component\Utility\Html;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\CloseDialogCommand;
use Drupal\Core\Ajax\AppendCommand;
use Drupal\Core\Ajax\OpenDialogCommand;
use Drupal\Core\Ajax\MessageCommand;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityListBuilder;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Form\FormInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Provides a listing of directories with enhanced UI features.
 */
class DirectoryListBuilder extends EntityListBuilder implements FormInterface {

  /**
   * The form builder.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;

  /**
   * The advanced file destination manager.
   *
   * @var \Drupal\advanced_file_destination\Service\AdvancedFileDestinationManager
   */
  protected $destinationManager;

  /**
   * The request stack service.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The date formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * {@inheritdoc}
   */
  protected $limit = 25;

  /**
   * Constructs a new DirectoryListBuilder object.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   * @param \Drupal\Core\Entity\EntityStorageInterface $storage
   *   The entity storage class.
   * @param \Drupal\Core\Form\FormBuilderInterface $form_builder
   *   The form builder.
   */
  public function __construct(EntityTypeInterface $entity_type, EntityStorageInterface $storage, FormBuilderInterface $form_builder, AdvancedFileDestinationManager $destination_manager, RequestStack $request_stack, ConfigFactoryInterface $config_factory, DateFormatterInterface $date_formatter) {
    parent::__construct($entity_type, $storage);
    $this->formBuilder = $form_builder;
    $this->destinationManager = $destination_manager;
    $this->requestStack = $request_stack;
    $this->configFactory = $config_factory;
    $this->dateFormatter = $date_formatter;
  }

  /**
   * {@inheritdoc}
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    return new static(
          $entity_type,
          $container->get('entity_type.manager')->getStorage($entity_type->id()),
          $container->get('form_builder'),
          $container->get('advanced_file_destination.manager'),
          $container->get('request_stack'),
          $container->get('config.factory'),
          $container->get('date.formatter')
      );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'directory_list_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // Add the library.
    $form['#attached']['library'][] = 'advanced_file_destination/directory_list';

    // Add filters.
    $form['filters'] = [
      '#type' => 'details',
      '#title' => $this->t('Filters'),
      '#open' => TRUE,
      '#attributes' => ['class' => ['directory-filters', 'clearfix']],
      'status' => [
        '#type' => 'select',
        '#title' => $this->t('Status'),
        '#options' => [
          '' => $this->t('- Any -'),
          '1' => $this->t('Active'),
          '0' => $this->t('Disabled'),
        ],
        '#default_value' => $this->requestStack->getCurrentRequest()->query->get('status'),
      ],

      'name' => [
        '#type' => 'search',
        '#title' => $this->t('Search by name'),
        '#default_value' => $this->requestStack->getCurrentRequest()->query->get('name'),
        '#size' => 30,
      ],
      'actions' => [
        '#type' => 'actions',
        'submit' => [
          '#type' => 'submit',
          '#value' => $this->t('Filter'),
          '#button_type' => 'primary',
        ],
        'reset' => [
          '#type' => 'link',
          '#title' => $this->t('Reset'),
          '#url' => Url::fromRoute('<current>'),
          '#attributes' => ['class' => ['button']],
        ],
      ],
    ];

    // Build table.
    $form['directories'] = [
      '#type' => 'details',
      '#title' => $this->t('Directories'),
      '#open' => TRUE,
      '#attributes' => ['class' => ['directory-listing']],
    ];
    $form['directories']['#prefix'] = '<div class="directory-listing-wrapper" id="directory-listing-wrapper">';
    $form['directories']['#suffix'] = '</div>';
    $form['directories']['table'] = [
      '#type' => 'table',
      '#header' => $this->buildHeader(),
      '#empty' => $this->t('No directories available.'),
      '#tableselect' => TRUE,
      '#attributes' => ['class' => ['directory-listing-table']],
    ];

    // Load entities with pagination.
    $entities = $this->load();
    // Sort entities by weight and then by name.
    usort($entities, function ($a, $b) {
            /**
              * @var \Drupal\advanced_file_destination\Entity\AdvancedFileDestinationDirectory $a
              * @var \Drupal\advanced_file_destination\Entity\AdvancedFileDestinationDirectory $b
            */
      // Compare by weight first.
      $weightA = method_exists($a, 'getWeight') ? $a->getWeight() : 0;
      $weightB = method_exists($b, 'getWeight') ? $b->getWeight() : 0;
      if ($weightA === $weightB) {
      // If weights are equal, compare by name.
      $nameA = method_exists($a, 'getName') ? $a->getName() : '';
      $nameB = method_exists($b, 'getName') ? $b->getName() : '';
      return strnatcasecmp($nameA, $nameB);
      }
      return $weightA <=> $weightB;
    });

    foreach ($entities as $entity) {
      $form['directories']['table'][$entity->id()] = $this->buildRow($entity);
    }

    // Add bulk operations with new structure.
    $form['directories']['operations'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => ['directory-bulk-operations'],
        'data-drupal-selector' => 'edit-bulk-actions',
        'role' => 'group',
      ],
      'status' => [
        '#type' => 'container',
        '#attributes' => [
          'class' => ['directory-bulk-operations__item', 'directory-bulk-operations__item--status'],
        ],
        '#markup' => $this->t('No items selected'),
      ],
      'action' => [
        '#type' => 'select',
        '#title' => $this->t('Action'),
        '#title_display' => 'invisible',
        '#options' => [
          'enable' => $this->t('Enable'),
          'disable' => $this->t('Disable'),
          'delete' => $this->t('Delete'),
        ],
        '#empty_option' => $this->t('- Select action -'),
        '#wrapper_attributes' => [
          'class' => ['directory-bulk-operations__item', 'directory-bulk-operations__item--preceding-actions'],
        ],
      ],
      'submit' => [
        '#type' => 'submit',
        '#value' => $this->t('Apply to selected items'),
        '#button_type' => 'primary',
        '#attributes' => ['class' => ['button--small']],
        '#name' => 'operations-submit',
        '#ajax' => [
          'callback' => '::submitBulkForm',
          'event' => 'click',
          'wrapper' => 'directory-listing-wrapper',
          'progress' => [
            'type' => 'throbber',
            'message' => $this->t('Processing...'),
          ],
        ],
      ],
    ];

    // Add pager.
    $form['pager'] = [
      '#type' => 'pager',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $filters = [];
    foreach (['status', 'name'] as $filter) {
      if ($value = $form_state->getValue($filter)) {
        $filters[$filter] = $value;
      }
    }
    if(!empty($filters)) {
      // If filters are set, add them to the URL.
      $form_state->setRedirect('<current>', $filters);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildHeader() {
    $header = [
      'preview' => [
        'data' => $this->t('Preview'),
        'class' => ['directory-preview-header'],
      ],
    ];

    $header += [
      'name' => [
        'data' => $this->t('Name'),
        'field' => 'label',
        'specifier' => 'label',
        'sort' => 'asc',
      ],
      'path' => [
        'data' => $this->t('Path'),
        'field' => 'path',
        'specifier' => 'path',
      ],
      'status' => [
        'data' => $this->t('Status'),
        'field' => 'status',
        'specifier' => 'status',
      ],
      'created' => [
        'data' => $this->t('Created'),
        'field' => 'created',
        'specifier' => 'created',
        'sort' => 'desc',
      ],
      'changed' => [
        'data' => $this->t('Updated'),
        'field' => 'changed',
        'specifier' => 'changed',
        'sort' => 'desc',
      ],
      'operations' => [
        'data' => $this->t('Operations'),
      ],
    ];

    return $header;
  }

  /**
   * {@inheritdoc}
   */
  public function load() {
    $query = $this->getStorage()->getQuery()
      ->accessCheck(TRUE)
      ->pager($this->limit);

    // Add filters.
    if ($status = $this->requestStack->getCurrentRequest()->query->get('status')) {
      $query->condition('status', $status);
    }
    if ($name = $this->requestStack->getCurrentRequest()->query->get('name')) {
      $query->condition('name', '%' . $name . '%', 'LIKE');
    }

    // Get sort parameters from URL or use defaults.
    $header = $this->buildHeader();
    $order = $this->requestStack->getCurrentRequest()->query->get('order');
    $sort = $this->requestStack->getCurrentRequest()->query->get('sort');

    // If no sort specified in URL, use changed/DESC as default.
    if (empty($sort)) {
      $sort = 'changed';
      $order = 'DESC';
    }

    // Ensure we have a valid sort field.
    if (!isset($header[$sort]['field'])) {
      $sort = 'changed';
      $order = 'DESC';
    }

    // Apply sorting.
    $query->sort($header[$sort]['field'], $order);

    $ids = $query->execute();
    return $this->storage->loadMultiple($ids);
  }

  /**
   * {@inheritdoc}
   */
  public function render() {
    return $this->formBuilder->getForm($this);
  }

  /**
   * Submit handler for bulk operations.
   */
  public function submitBulkForm(array &$form, FormStateInterface $form_state) {
    $response = new AjaxResponse();

    $action = $form_state->getValue( 'action');
    if (empty($action)) {
      $response->addCommand(new MessageCommand($this->t('Please select an action to perform.'), NULL, ['type' => 'error']));
      return $response;
    }
    $selected = array_filter($form_state->getValue('table', []));
    if (empty($selected)) {
      $response->addCommand(new MessageCommand($this->t('No directories selected for the action.'), NULL, ['type' => 'error']));
      return $response;
    }

    // Display a confirmation dialog for bulk actions.
    $dialog_title = $this->t('Confirm Bulk Action');
    $dialog_message = '<p>'. $this->t('Are you sure you want to <b>%action</b> the selected directories?', ['%action' => ucfirst($action)]) . '</p>';
    if ($action === 'delete') {
      $dialog_message .= '<p><b>' . $this->t('This action cannot be undone.') . '</b></p>';
    }
    else {
      $dialog_message .= '<p>' . $this->t('This action can be reversed later.') . '</p>';
    }
    $dialog_message .= '<p>' . $this->t('Selected directories: @count', ['@count' => count($selected)]) . '</p>';
    $dialog_message .= '<ul>';
    foreach ($selected as $id => $value) {
      if ($value) {
        $entity = $this->getStorage()->load($id);
        if ($entity) {
              /**
                * @var \Drupal\advanced_file_destination\Entity\AdvancedFileDestinationDirectory $entity
              */
          $dialog_message .= '<li>' . Html::escape($entity->getName()) . ' (' . Html::escape($entity->getPath()) . ')</li>';
        }
      }
    }
    $dialog_message .= '</ul>';
    $dialog_message .= '<p>' . $this->t('Do you want to proceed?') . '</p>';

    $dialog = [
      '#type' => 'form',
      '#id' => 'directory-listing-modal',
      '#attributes' => ['class' => ['directory-bulk-confirmation', 'dialog-content', 'clearfix']],
      'message' => [
        '#markup' => $dialog_message,
      ],
      'action_hidden' => [
        '#type' => 'hidden',
        '#value' => $action,
        '#name' => 'action',
      ],
      'selected_hidden' => [
        '#type' => 'hidden',
        '#value' => json_encode(array_keys($selected)),
        '#name' => 'selected',
      ],
      'actions' => [
        '#type' => 'conatainer',
        '#attributes' => ['class' => ['directory-bulk-confirmation-actions']],
        'confirm' => [
          '#type' => 'button',
          '#value' => $this->t('Confirm'),
          '#button_type' => 'primary',
          '#ajax' => [
            'callback' => [$this, 'executeBulkAction'],
            'event' => 'click',
            'wrapper' => 'directory-listing-wrapper',
            'progress' => [
              'type' => 'throbber',
              'message' => $this->t('Processing...'),
            ],
          ],
        ],
        'cancel' => [
          '#type' => 'link',
          '#title' => $this->t('Cancel'),
          '#url' => Url::fromRoute('<current>'),
          '#attributes' => ['class' => ['button']],
        ],
      ],
    ];
    $dialog_title = $this->t('Bulk Action Confirmation');
    $response->addCommand(new OpenDialogCommand('#directory-listing-modal', $dialog_title, $dialog, [
      'width' => '600px',
      'closeOnEscape' => TRUE,
      'dialogClass' => 'advanced-file-destination-modal',
      'modal' => TRUE,
    ]));
    return $response;
  }

  /**
   * Executes the selected bulk action.
   */
  public function executeBulkAction(array &$form, FormStateInterface $form_state): AjaxResponse {
    // debug log
    \Drupal::logger('advanced_file_destination')->debug('Executing bulk action: @action with selected directories: @selected', [
      '@action' => $form_state->getValue('action_hidden'),
      '@selected' => $form_state->getValue('selected_hidden'),
    ]);
    $action = $form_state->getValue('action_hidden');
    $selected = json_decode($form_state->getValue('selected_hidden'), TRUE);
    $response = new AjaxResponse();

    if (empty($action)) {
      $response->addCommand(new MessageCommand($this->t('No action selected.'), NULL, ['type' => 'error']));
      return $response;
    }

    if (empty($selected)) {
      $response->addCommand(new MessageCommand($this->t('No directories selected.'), NULL, ['type' => 'error']));
      return $response;
    }

    // Perform the action.
    try {
      foreach ($selected as $id) {
        // The $selected array from JSON decode might not be in the format $id => $value
        // if it was just an array of IDs. Assuming $id is the entity ID directly.
        $entity = $this->getStorage()->load($id);
        if ($entity) {
          /**
            * @var \Drupal\advanced_file_destination\Entity\AdvancedFileDestinationDirectory $entity
          */
          switch ($action) {
            case 'enable':
              $entity->setStatus(TRUE);
              break;

            case 'disable':
              $entity->setStatus(FALSE);
              break;

            case 'delete':
              $entity->delete();
              // Skip saving for delete, as it's already handled by delete().
              continue 2;

            default:
              throw new \InvalidArgumentException($this->t('Unknown action: %action', ['%action' => $action]));
          }
          $entity->save();
        }
      }

      $response->addCommand(new CloseDialogCommand('#directory-listing-modal'));
      $response->addCommand(new MessageCommand($this->t('Successfully performed the %action action on @count directories.', [
        '%action' => $action,
        '@count' => count($selected),
      ])));
      // Re-render the form to get the updated list.
      $updated_form = $this->render();
      $response->addCommand(new ReplaceCommand('#directory-listing-wrapper', $updated_form['directories']));
      return $response;
    }
    catch (\Exception $e) {
      $response->addCommand(new CloseDialogCommand('#directory-listing-modal'));
      $response->addCommand(new MessageCommand($this->t('An error occurred while performing the bulk action: %message', ['%message' => $e->getMessage()]), NULL, ['type' => 'error']));
      return $response;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    if ($form_state->getTriggeringElement()['#parents'][0] === 'operations') {
      // Validate bulk operations.
      $selected = array_filter($form_state->getValue('table', []));
      if (empty($selected)) {
        $form_state->setErrorByName('operations', $this->t('No items selected.'));
      }

      $action = $form_state->getValue(['operations', 'action']);
      if (empty($action)) {
        $form_state->setErrorByName('operations[action]', $this->t('Please select an operation.'));
      }
    }
    else {
      // Validate filters.
      if ($name = $form_state->getValue('name')) {
        if (strlen($name) < 2) {
          $form_state->setErrorByName('name', $this->t('Search term must be at least 2 characters.'));
        }
      }
    }
  }

  /**
   * Builds a row for the directory list.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @return array
   *   A render array representing the row.
   */
  public function buildRow(EntityInterface $entity) {
    /**
      * @var \Drupal\advanced_file_destination\Entity\AdvancedFileDestinationDirectory $entity
    */
    $row = [];

    if ($this->containsNonPublicScheme($entity->getPath())) {
      return;
    }

    // Add preview column with folder icon.
    $row['preview'] = [
      'data' => [
        '#type' => 'container',
        '#attributes' => ['class' => ['directory-preview-wrapper']],
        'icon' => [
          '#type' => 'html_tag',
          '#tag' => 'span',
          '#attributes' => [
            'class' => ['directory-icon'],
            'title' => $this->t('Directory'),
          ],
        ],
      ],
    ];

    // Check if path exists and is a directory.
    $path = $entity->getPath();
    $is_directory = file_exists($path) && is_dir($path);

    $row['name'] = [
      'data' => [
        '#type' => 'container',
        '#attributes' => ['class' => ['directory-name-wrapper']],
        'icon' => [
          '#type' => 'html_tag',
          '#tag' => 'span',
          '#attributes' => [
            'class' => ['directory-icon'],
            'title' => $this->t('Directory'),
          ],
        ],
        'link' => [
          '#type' => 'link',
          '#title' => $entity->getName(),
          '#url' => $is_directory
            ? Url::fromRoute('entity.afd_directory.edit_form', ['afd_directory' => $entity->id()])
            : $entity->toUrl('edit-form'),
        ],
      ],
    ];

    $row['path'] = [
      'data' => ['#markup' => $path],
    ];

    $row['status'] = [
      'data' => [
        '#markup' => $entity->isActive() ? $this->t('Active') : $this->t('Disabled'),
        '#wrapper_attributes' => [
          'class' => [
            'directory-status', $entity->isActive() ?
            'directory-status--active' :
            'directory-status--disabled',
          ],
        ],
      ],
    ];

    $row['created'] = [
      'data' => [
        '#markup' => $this->dateFormatter->format($entity->getCreatedTime()),
      ],
    ];

    $row['changed'] = [
      'data' => [
        '#markup' => $entity->getChangedTime() ? $this->dateFormatter->format($entity->getChangedTime()) : $this->dateFormatter->format($entity->getCreatedTime()),
      ],
    ];

    $row['operations'] = [
      'data' => $this->buildOperations($entity),
    ];

    return $row;
  }

  /**
   * Checks if the path contains a non-public scheme.
   *
   * @param string $path
   *   The path to check.
   *
   * @return bool
   *   TRUE if the path contains a non-public scheme, FALSE otherwise.
   */
  protected function containsNonPublicScheme($path) {
    return $this->destinationManager->containsNonPublicScheme($path);
  }

}
